import multiprocessing
import os
from itertools import product

import numpy as np
from joblib import Parallel, delayed, parallel_backend

from manifoldy.utils import get_instance_name
from manifoldy.generators import gen_pair
from manifoldy.L2NormCurvature import L2_norm_sectional_curvature
from manifoldy.definitions import (
    RANDOM_NOISE_STD,
    TARGET_DIMENSIONALITY,
    USED_CURVATURES,
    DIFFICULTY,
    DIMENSIONALITY_REDUCTION_MODELS,
)
from manifoldy.test.generators_before_refactor import (
    create_dataset as create_dataset_refactor,
)
from manifoldy.test.L2NormCurvature_before_refactor import (
    L2_norm_sectional_curvature as L2_norm_sectional_curvature_refactor,
)

grid = np.mgrid[0:1:30j, 0:1:30j].reshape(2, -1).T


def gen_pair_refactor(instance):
    os.system(
        "taskset -cp 0-%d %s > /dev/null 2>&1"
        % (multiprocessing.cpu_count(), os.getpid())
    )
    name = get_instance_name(instance)
    return name, np.apply_along_axis(
        create_dataset_refactor(
            [instance[0], instance[1]],
            (instance[2], instance[3]),
            TARGET_DIMENSIONALITY,
            RANDOM_NOISE_STD,
        ),
        1,
        grid,
    )


def aux_test_gen_pair(instance):
    name, data = gen_pair(instance)
    name_r, data_r = gen_pair_refactor(instance)
    print(type(name), type(name_r))
    assert name == name_r, (name, name_r)
    print(type(data), type(data_r))
    assert np.array_equal(data, data_r), (name, name_r, data, data_r)


def test_gen_pair():
    with parallel_backend("loky"):
        print("Generating problem instances...")
        Parallel(n_jobs=-1)(
            delayed(aux_test_gen_pair)(x)
            for x in product(USED_CURVATURES, USED_CURVATURES, DIFFICULTY, DIFFICULTY)
        )


def eval_pair(p, model):
    os.system(
        "taskset -cp 0-%d %s > /dev/null 2>&1"
        % (multiprocessing.cpu_count(), os.getpid())
    )
    return (
        p[0],
        type(model).__name__,
        L2_norm_sectional_curvature(
            grid,
            model.fit_transform(p[1]),
            metric_estimation="interpolate_metric",
            verbose=False,
        ),
    )


def eval_pair_refactor(p, model):
    os.system(
        "taskset -cp 0-%d %s > /dev/null 2>&1"
        % (multiprocessing.cpu_count(), os.getpid())
    )
    return (
        p[0],
        type(model).__name__,
        L2_norm_sectional_curvature_refactor(
            grid,
            model.fit_transform(p[1]),
            metric_estimation="interpolate_metric",
            verbose=False,
        ),
    )


def aux_test_eval_pair(p, model):
    _, _, original = eval_pair(p, model)
    _, _, refactor = eval_pair_refactor(p, model)
    print(p[0], type(model).__name__, original, refactor, original == refactor)
    assert original == refactor


def test_eval_pair():
    with parallel_backend("loky"):
        print("Generating problem instances...")
        results = Parallel(n_jobs=-1)(
            delayed(gen_pair)(x)
            for x in product(USED_CURVATURES, USED_CURVATURES, DIFFICULTY, DIFFICULTY)
        )
        instance = dict(results)
        print("Evaluating")
        Parallel(n_jobs=-1)(
            delayed(aux_test_eval_pair)(x, y)
            for x, y in product(instance.items(), DIMENSIONALITY_REDUCTION_MODELS)
        )



import scipy as sc
import numpy as np
from scipy.interpolate import CubicSpline
from scipy.stats import multivariate_normal
from scipy.stats import special_ortho_group

# Not part of the original, but needed for reproducibility and for tests.
from manifoldy.definitions import RANDOM_SEED


def prod_complex(z1, z2):
    return z1[0] * z2[0] - z1[1] * z2[1], z1[0] * z2[1] + z1[1] * z2[0]


def my_nintegration(f, a, b):
    return sc.integrate.quad(f, a, b)[0]


def create_curve(k, x0, v0, t_min=-1, t_max=1, t_step=0.01):
    t_num = (t_max - t_min) / t_step
    xis = np.linspace(t_min, t_max, 10 * int(t_num))
    theta_fun = np.vectorize(lambda x: my_nintegration(k, 0, x))
    theta_val = theta_fun(xis)
    sp_theta = CubicSpline(xis, theta_val)

    def _tang(xi):
        thet = sp_theta(xi)
        return prod_complex(v0, [np.cos(thet), np.sin(thet)])

    def _tang_x(xi):
        return _tang(xi)[0]

    def _tang_y(xi):
        return _tang(xi)[1]

    si = np.linspace(t_min, t_max, 10 * int(t_num))
    res_x_fun = np.vectorize(lambda x: my_nintegration(_tang_x, 0, x) + x0[0])
    res_y_fun = np.vectorize(lambda x: my_nintegration(_tang_y, 0, x) + x0[1])
    res_x_val = res_x_fun(si)
    res_y_val = res_y_fun(si)

    sp_res_x = CubicSpline(si, res_x_val)
    sp_res_y = CubicSpline(si, res_y_val)

    def _gamma(s):
        return np.array([sp_res_x(s), sp_res_y(s)])

    return _gamma


def create_spatial_curve(k, x0, v0, n, i, t_min=-1, t_max=1, t_step=0.01):
    gamma = create_curve(k, x0, v0, t_min, t_max, t_step)

    def _gamma_sp(s):
        return np.array([0] * (i) + list(gamma(s)) + [0] * (n - i - 2))

    return _gamma_sp


def create_submanifold(k, x0, v0, n, t_min=-1, t_max=1, t_step=0.01):
    d = len(k)

    if d > n - 1:
        print("Too big submanifold")

    if len(x0) != d or len(v0) != d:
        print("Dimensions do not match!")

    gammas = []
    for i in range(d):
        gammas.append(
            create_spatial_curve(k[i], x0[i], v0[i], n, i, t_min, t_max, t_step)
        )

    def _X(arg):
        suma = 0
        for i in range(d):
            suma += gammas[i](arg[i])
        return suma

    return _X


def relu(t, slope):
    if t < 0:
        return 0
    return slope * t


def get_curvature_type(curvature_type, args):
    if curvature_type == "flat":
        k = lambda t: 0
    elif curvature_type == "circle":
        k = lambda t: 2 * np.pi * (args)
    elif curvature_type == "polynomial_roll":
        k = lambda t: 4 * args * (t + 1) ** (2 * args)
    elif curvature_type == "roll":
        k = lambda t: np.exp(4 * args * t)
    elif curvature_type == "gaussian":
        k = lambda t: 1 / (0.1 * args) * np.exp(-((t) ** (2)) / (0.1 * args) ** 2)
    elif curvature_type == "sine":
        k = lambda t: (5 + (args - 1) * 10) * np.sin(2 * np.pi * t)
    elif curvature_type == "logistic":
        k = lambda t: (10 * args) / (1 + np.exp(-0.5 * t))
    elif curvature_type == "relu":
        k = lambda t: relu(t, 10 * args)

    return k


def create_curve_type(curvature_type, *args):
    return create_curve(get_curvature_type(curvature_type, *args), (0, 0), (1, 0))


def create_submanifold_type(curvature_types, args, n):
    d = len(curvature_types)
    k = [get_curvature_type(curvature_types[i], args[i]) for i in range(d)]
    if any(ks is None for ks in k):
        return None

    x0 = [(0, 0)] * d
    v0 = [(1, 1)] * d

    return create_submanifold(k, x0, v0, n)


def create_dataset_anisotropic(curvature_types, args, n, cov):
    X = create_submanifold_type(curvature_types, args, n)
    R = special_ortho_group.rvs(n, random_state=RANDOM_SEED)
    if X is None:
        print("Error during creation of the submanifold")

    def _phi(*args):
        return np.dot(
            R,
            X(*args)
            + multivariate_normal.rvs(
                mean=[0] * n, cov=cov, random_state=RANDOM_SEED
            ).T,
        ).T

    return _phi


def create_dataset(curvature_types, args, n, std):
    return create_dataset_anisotropic(
        curvature_types, args, n, np.identity(n) * std ** 2
    )



#!/usr/bin/env python
# coding: utf-8

import multiprocessing
import os
import pickle
from itertools import product
import matplotlib.pyplot as plt
import numpy as np

from joblib import Parallel, delayed, parallel_backend

from definitions import DIMENSIONALITY_REDUCTION_MODELS
from utils import setup_multiprocessing


def apply_reduction(instance, model):
    name, instance_data = instance
    os.system(
        "taskset -cp 0-%d %s > /dev/null 2>&1"
        % (multiprocessing.cpu_count(), os.getpid())
    )
    return name, type(model).__name__, model.fit_transform(instance_data)


if __name__ == "__main__":
    setup_multiprocessing()

    with open("results/dataset_names.pickle", "rb") as file:
        names = pickle.load(file)

    data = np.load("results/dataset.npy")

    instances = zip(names, data)

    with parallel_backend("loky"):
        print("Applying dimensionality reduction algorithms...")
        results = Parallel(n_jobs=-1)(
            delayed(apply_reduction)(instance, model)
            for instance, model in product(instances, DIMENSIONALITY_REDUCTION_MODELS)
        )

        projection_names = [x[0] + " " + x[1] for x in results]
        projection_data = [x[2] for x in results]

        with open("results/projection_dataset_names.pickle", "wb") as file:
            pickle.dump(projection_names, file)
        np.save("results/projection_results.npy", projection_data)


i = 0
plt.scatter(projection_data[i][:, 0], projection_data[i][:, 1])

for i in range(10):
    print(projection_names[i])
    plt.scatter(projection_data[i][:, 0], projection_data[i][:, 1])
    plt.show()



#!/usr/bin/env python
# coding: utf-8

import copy
import datetime
import multiprocessing
import os
import pickle
from functools import reduce

import ndsplines
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from joblib import parallel_backend, Parallel, delayed
from scipy.integrate import nquad
from scipy.interpolate import griddata
from scipy.spatial.distance import cdist
from scipy.special import comb

from definitions import GRID


def compute_first_christoffel_symbols(dg, n):
    CS_first = np.zeros((n, n, n))
    for i in range(n):
        for j in range(n):
            for k in range(n):
                CS_first[k, i, j] = 1 / 2 * (dg[j][k, i] + dg[i][k, j] - dg[k][i, j])

    return CS_first


def compute_partial_derivative_first_christoffel_symbols(ddg, l, n):
    CS_first_diff = np.zeros((n, n, n))
    for i in range(n):
        for j in range(n):
            for k in range(n):
                CS_first_diff[k, i, j] = (
                    1 / 2 * (ddg[l][j][k, i] + ddg[l][i][k, j] - ddg[l][k][i, j])
                )

    return CS_first_diff


def compute_derivative_first_christoffel_symbols(ddg, n):
    derivatives = []
    for l in range(n):
        CS_first_diff = compute_partial_derivative_first_christoffel_symbols(ddg, l, n)
        derivatives.append(CS_first_diff)
    return np.stack(derivatives)


def compute_christoffel_symbols(g, dg):
    # Index 0: Upper index
    # Index 1, 2: Lower indices
    # Gamma_{1,2}^0
    n = g.shape[0]
    CS_first = compute_first_christoffel_symbols(dg, n)

    g_inv = np.linalg.inv(g)
    return np.tensordot(g_inv, CS_first, axes=([1], [0]))


def compute_derivative_christoffel_symbols(g, dg, ddg):
    # Index 0: Derivative of the symbol
    # Index 1: Upper index
    # Index 2, 3: Lower indices
    # Gamma_{0;2,3}^1
    n = g.shape[0]
    CS_first = compute_first_christoffel_symbols(dg, n)
    g_inv = np.linalg.inv(g)

    derivatives = []
    for l in range(n):
        CS_first_diff_l = compute_partial_derivative_first_christoffel_symbols(
            ddg, l, n
        )
        g_inv_diff_l = -np.matmul(np.matmul(g_inv, dg[l]), g_inv)
        CS_l = np.tensordot(g_inv_diff_l, CS_first, axes=([1], [0])) + np.tensordot(
            g_inv, CS_first_diff_l, axes=([1], [0])
        )
        derivatives.append(CS_l)

    return np.stack(derivatives)


def compute_curvature_tensor(g, dg, ddg):
    # Index 0: Upper index
    # Index 1, 2, 3: Lower indices
    # R[l,i,j,k] = R^l_{ijk} = dx^l(R(ei,ej)ek)
    n = g.shape[0]

    CS = compute_christoffel_symbols(g, dg)
    CS_diff = compute_derivative_christoffel_symbols(g, dg, ddg)
    R = np.zeros((n, n, n, n))
    CS_contract = np.tensordot(CS, CS, axes=([0], [2]))

    for i in range(n):
        for j in range(n):
            for k in range(n):
                for l in range(n):
                    R[l, i, j, k] = (
                        CS_diff[j, l, i, k]
                        - CS_diff[i, l, j, k]
                        + CS_contract[i, k, l, j]
                        - CS_contract[j, k, l, i]
                    )

    return R


def compute_first_curvature_tensor(g, dg, ddg):
    # Index 0: Upper index
    # Index 1, 2, 3: Lower indices
    # R[l,i,j,k] = R^l_{ijk} = dx^l(R(ei,ej)ek)
    # n = g.shape[0]

    R = compute_curvature_tensor(g, dg, ddg)
    R_first = np.tensordot(R, g, axes=([0], [0]))

    return R_first


def compute_sectional_curvatures(g, dg, ddg):
    # Returns a vector whose entries are
    # sectional curvatures at (e1,e2), (e1,e3),...,(e1,en)
    # (e2,e3), (e2,e4), ..., (e2, en),...
    n = g.shape[0]
    R = compute_first_curvature_tensor(g, dg, ddg)
    sect_curvs = np.zeros(int(comb(n, 2)))
    counter = 0
    for i in range(n):
        for j in range(i + 1, n):
            area = g[i, i] * g[j, j] - g[i, j] ** 2
            sect_curvs[counter] = R[i, j, i, j] / area
            counter += 1

    return np.stack(sect_curvs)


# UTILS


def dataset_to_grid(X, n_samples, dim):
    shap = n_samples + [dim]
    # shap = [dim] + n_samples
    return X.reshape(shap)


# +
N_SAMPLES_GRID = 41


def grid_data(X, Y, n_samples=None):
    dim = X.shape[1]
    if n_samples is None:
        n_samples = [N_SAMPLES_GRID] * dim

    offset = 0.05
    xs = np.array(
        [np.linspace(0 + offset, 1 - offset, n_samples[i]) for i in range(dim)]
    )
    X_new = np.array(np.meshgrid(*xs))

    Y_new = []
    for i in range(dim):
        grid_y = griddata(X, Y[:, i], tuple(X_new), method="linear")
        Y_new.append(grid_y)
    Y_new = np.stack(Y_new)

    n_total_samples = 1
    for n in n_samples:
        n_total_samples *= n

    return (
        X_new.reshape(dim, n_total_samples).T,
        Y_new.reshape(dim, n_total_samples).T,
        n_samples,
    )


# -


def estimate_inner_product(X, b):
    u, s, vh = np.linalg.svd(X, full_matrices=True)
    S = np.zeros((u.shape[0], vh.shape[0]))
    S_inv = np.zeros((vh.shape[0], u.shape[0]))
    np.fill_diagonal(S, s, wrap=False)
    np.fill_diagonal(S_inv, 1 / s, wrap=False)

    return vh.T @ (S_inv @ (u.T @ (b @ (u @ (S_inv.T @ vh)))))


def estimate_tensor_left(X, b, arity):
    X_inv = np.linalg.pinv(X)
    result = b
    for _ in range(arity):
        result = np.tensordot(X_inv, result, axes=([1], [0]))

    return result


def quadrature_regular_grid(X, values, grid_n_samples):
    x0 = X[0]
    index = 1
    volume = 1
    n_samples = copy.copy(grid_n_samples)
    aux = n_samples[0]
    n_samples[0] = n_samples[1]
    n_samples[1] = aux
    n_samples = [1] + n_samples[:-1]
    for i in n_samples:
        index = index * i
        volume = volume * np.linalg.norm(X[index] - x0)

    return np.sum(values, axis=0) * volume


def crop_matrix(X, crop_indices):
    if len(X.shape) == 1:
        return X[crop_indices[0]]

    return np.stack([crop_matrix(f, crop_indices[1:]) for f in X[crop_indices[0]]])


# ESTIMATION OF THE RIEMANNIAN METRIC


def nearest_neighbors_indices(i, n_neighbors, distance_matrix):

    indices = np.argsort(distance_matrix[i], axis=0)[: n_neighbors + 1]
    indices = np.delete(indices, 0, axis=0)

    return indices


def get_riemannian_metric(X, Y, n_neighbors, distance_matrix):
    g = []
    dim = X.shape[1]
    for i in range(X.shape[0]):
        stop = False
        k = n_neighbors
        while not stop:
            index_neig = nearest_neighbors_indices(i, k, distance_matrix)
            X_extract = X[index_neig]
            X_vec = X_extract - X[i]
            if np.linalg.matrix_rank(X_vec) >= dim:
                stop = True
            else:
                k += 1

        Y_extract = Y[index_neig]
        Y_vec = Y_extract - Y[i]
        b = np.matmul(Y_vec, Y_vec.T)

        g.append(estimate_inner_product(X_vec, b))
    g = np.stack(g)
    return g


def compute_knn_point_derivatives(X, Y, i, n_neighbors, distance_matrix):
    stop = False
    k = n_neighbors
    dim = X.shape[1]
    while not stop:
        index_neig = nearest_neighbors_indices(i, k, distance_matrix)
        X_extract = X[index_neig]
        X_vec = X_extract - X[i]
        if np.linalg.matrix_rank(X_vec) >= dim:
            stop = True
        else:
            k += 1

    # new_shape = (k,) + Y.shape[1:]
    # b = np.zeros(new_shape)
    Y_extract = Y[index_neig]
    Y_vec = Y_extract - Y[i]

    return estimate_tensor_left(X_vec, Y_vec, 1)


def compute_knn_derivative(X, Y, n_neighbors, distance_matrix):
    der = []
    for i in range(X.shape[0]):
        der.append(compute_knn_point_derivatives(X, Y, i, n_neighbors, distance_matrix))

    return np.stack(der)


def get_knn_derivatives(X, Y, n_neighbors, distance_matrix):
    first = compute_knn_derivative(X, Y, n_neighbors, distance_matrix)
    second = compute_knn_derivative(X, first, n_neighbors, distance_matrix)
    third = compute_knn_derivative(X, second, n_neighbors, distance_matrix)

    return first, second, third


def get_partial_derivative_indices(variables, order=2):
    """
    Gets indices for partial derivatives w.r.t each variable.
    :param variables: number of variables (number of partial derivatives per order)
    :param order: max. order of derivation.
    :return: list of length order, containing permutations of every partial derivative that can be taken for
    every variable starting from the indices of the previous order.
    """
    first_order_indices = np.eye(variables, dtype=int)
    ret = [first_order_indices]

    previous_order_indices = first_order_indices
    for _ in range(1, order):
        current_order_indices = []
        for previous_order_index in previous_order_indices:
            for variable in range(variables):
                partial_derivative_index = list(previous_order_index)
                partial_derivative_index[variable] += 1
                current_order_indices.append(partial_derivative_index)

        previous_order_indices = np.array(current_order_indices)
        ret.append(previous_order_indices)

    return ret


def interpolate_derivative_function(X, Y, order=3):
    dim = Y.shape[1]

    f = []
    for i in range(dim):
        tidy_data = np.hstack((X, Y[:, i].reshape(-1, 1)))
        interp = ndsplines.make_interp_spline_from_tidy(tidy_data, range(dim), [dim])
        f.append(interp)

    partial_derivative_indices = get_partial_derivative_indices(dim, order)

    derivatives = []
    for derivative_order in range(order):
        derivatives_list = []
        for indices in partial_derivative_indices[derivative_order]:
            derivatives_list.append(
                np.stack([f[j](X, nus=indices) for j in range(dim)])
            )

        derivatives_list = np.stack(derivatives_list)
        derivatives_list = derivatives_list.reshape(
            *tuple([dim] * (derivative_order + 2)), len(X)
        )

        derivatives_list = np.moveaxis(
            derivatives_list,
            range(derivative_order + 3),
            [(x + 1) % (derivative_order + 3) for x in range(derivative_order + 3)],
        )
        derivatives.append(derivatives_list)

    return tuple(derivatives)


def interpolate_metric_derivatives(g_sp, X, n_samples, order=2):
    n = reduce(lambda x, y: x * y, n_samples)
    d = len(n_samples)

    grid = dataset_to_grid(X, n_samples, X.shape[1])

    g = np.array([g_sp[i](grid) for i in range(d ** 2)])
    g = g.T.reshape((n, d, d))

    partial_derivative_indices = get_partial_derivative_indices(d, order=order)

    derivatives = []
    for derivative_order in range(order):
        derivatives_list = []
        for indices in partial_derivative_indices[derivative_order]:
            derivatives_list.append(
                np.stack([g_sp[i](grid, nus=indices) for i in range(d ** 2)])
            )

        derivatives_list = np.array(derivatives_list).reshape(
            *tuple([d] * (derivative_order + 3)), n
        )

        derivatives_list = np.moveaxis(
            derivatives_list,
            range(derivative_order + 4),
            [(x + 1) % (derivative_order + 4) for x in range(derivative_order + 4)],
        )
        derivatives.append(derivatives_list)

    return g, *derivatives


def interpolate_metric(X, Y, n_neighbors, distance_matrix):
    n = X.shape[0]
    d = X.shape[1]
    g = get_riemannian_metric(X, Y, n_neighbors, distance_matrix)
    inter = []
    for i in range(d):
        for j in range(d):
            tidy_data = np.hstack((X, g[:, i, j].reshape(n, 1)))
            interp = ndsplines.make_interp_spline_from_tidy(
                tidy_data,
                range(d),  # columns to use as independent variable data
                [d],  # columns to use as dependent variable data
            )
            inter.append(interp)

    return inter


# TOP FUNCTIONS


def combine_derivatives(df, ddf, dddf):
    g = np.einsum("ijs,iks->ijk", df, df, optimize="greedy")
    dg = np.einsum("ijks,ils->ijkl", ddf, df) + np.einsum(
        "iks,ijls->ijkl", df, ddf, optimize="greedy"
    )
    ddg = (
        np.einsum("ijkls,ims->ijklm", dddf, df, optimize="greedy")
        + np.einsum("ikls,ijms->ijklm", ddf, ddf, optimize="greedy")
        + np.einsum("ijls,ikms->ijklm", ddf, ddf, optimize="greedy")
        + np.einsum("ils,ijkms->ijklm", df, dddf, optimize="greedy")
    )
    return g, dg, ddg


def compute_distance_matrix(X, distance_metric=None):
    if distance_metric is None:

        def distance_metric(x, y):
            return np.linalg.norm(x - y)

    return cdist(X, X, metric=distance_metric)


def estimate_sectional_curvature(
    X, Y, n_neighbors, grid_n_samples, metric_estimation="KNN"
):

    if metric_estimation == "interpolate_metric":
        distance_matrix = compute_distance_matrix(X)
        g_sp = interpolate_metric(X, Y, n_neighbors, distance_matrix)
        g, dg, ddg = interpolate_metric_derivatives(g_sp, X, grid_n_samples)
    elif metric_estimation == "KNN":
        distance_matrix = compute_distance_matrix(X)
        df, ddf, dddf = get_knn_derivatives(X, Y, n_neighbors, distance_matrix)
        g, dg, ddg = combine_derivatives(df, ddf, dddf)
    elif metric_estimation == "interpolate":
        df, ddf, dddf = interpolate_derivative_function(X, Y)
        g, dg, ddg = combine_derivatives(df, ddf, dddf)
    else:
        raise ValueError(
            f'Accepted values for param metric_estimation are {{"KNN", "interpolate", "interpolate_metric"}} ('
            f"received {metric_estimation})"
        )

    R = []
    for i in range(len(X)):
        R.append(compute_sectional_curvatures(g[i], dg[i], ddg[i]))
    return np.stack(R)


def compute_grid_norm(X, R_sq, grid_n_samples, offset=2, verbose=False):
    """
    If offset < 1, the integration is carried out on the square [offset, 1-offset]^d.
    If offset >= 1, the integration is carried out by pulling appart offset points at
    each side of the cube (preferred option)
    """

    dim = X.shape[1]

    if offset < 0.1:
        reduction_indices = [
            range(
                int(grid_n_samples[i] * offset), int(grid_n_samples[i] * (1 - offset))
            )
            for i in range(len(grid_n_samples))
        ]

    else:
        reduction_indices = [
            range(offset, grid_n_samples[i] - offset)
            for i in range(len(grid_n_samples))
        ]

    X_grid = dataset_to_grid(X, grid_n_samples, dim)
    X_red_indices = reduction_indices + [range(dim)]
    X_crop_grid = crop_matrix(X_grid, X_red_indices)

    R_sq_grid = dataset_to_grid(R_sq, grid_n_samples, R_sq.shape[1])
    R_sq_red_indices = reduction_indices + [range(R_sq.shape[1])]
    R_sq_crop_grid = crop_matrix(R_sq_grid, R_sq_red_indices)

    n_crop = 1
    grid_n_samples_crop = [len(r) for r in reduction_indices]
    for r in grid_n_samples_crop:
        n_crop = n_crop * r
    X_crop = X_crop_grid.reshape(n_crop, X.shape[1])
    R_sq_crop = R_sq_crop_grid.reshape(n_crop, R_sq.shape[1])

    if verbose:
        print("Initiating grid integration")

    return np.sqrt(
        quadrature_regular_grid(X_crop, R_sq_crop, grid_n_samples_crop).sum()
    )


def compute_adaptive_norm(X, R_sq, offset, verbose):
    dim = X.shape[1]
    result = 0
    for i in range(R_sq.shape[1]):
        tidy_data = np.hstack((X, R_sq[:, i].reshape(X.shape[0], 1)))
        interp = ndsplines.make_interp_spline_from_tidy(
            tidy_data,
            range(dim),  # columns to use as independent variable data
            [dim],  # columns to use as dependent variable data
        )

        def f(*argv):
            return interp(argv)[0, 0]

        if verbose:
            print("Initiating adaptive integration")
        result += nquad(f, [[offset, 1 - offset]] * dim, full_output=False)[0]
    return np.sqrt(result)


def L2_norm_sectional_curvature(
    X,
    Y,
    metric_estimation="KNN",
    n_neighbors=20,
    integration="grid",
    grid_n_samples=None,
    offset=2,
    verbose=False,
    full_output=False,
):
    """
    Available metric_estimation:
        * 'KNN': Computes metric derivatives using n_neighbors nearest neighbors.
        * 'interpolate': Interpolates the underlying function and uses it to compute the metric.
        * 'interpolate_metric': Computes the metric from the data and interpolate it to compute derivatives.

    Available integrations:
        * 'grid': Computes the integral using a grid.
        * 'adaptive': Computes the integral using an adaptive method.
    """

    if integration not in {"adaptive", "grid"}:
        raise ValueError(
            f'Accepted values for param integration are {{"adaptive", "grid"}} (received {integration})'
        )

    if metric_estimation not in {"KNN", "interpolate", "interpolate_metric"}:
        raise ValueError(
            f'Accepted values for param metric_estimation are {{"KNN", "interpolate", '
            f'"interpolate_metric"}} (received {metric_estimation})'
        )

    if integration == "grid" and grid_n_samples is None:
        X, Y, grid_n_samples = grid_data(X, Y)

    if verbose:
        print("Starting: estimation of sectional curvature")
    R_sq = estimate_sectional_curvature(
        X, Y, n_neighbors, grid_n_samples, metric_estimation
    )
    if verbose:
        print("Estimation of sectional curvature: DONE")

    R_sq_log = np.log(1 + np.abs(R_sq))
    if integration == "adaptive":
        if full_output:
            return X, Y, R_sq_log, compute_adaptive_norm(X, R_sq_log, offset, verbose)
        return compute_adaptive_norm(X, R_sq_log, offset, verbose)

    if full_output:
        return (
            X,
            Y,
            R_sq_log,
            compute_grid_norm(X, R_sq_log, grid_n_samples, offset, verbose),
        )
    return compute_grid_norm(X, R_sq_log, grid_n_samples, offset, verbose)


# +


def plot_results(dataframe):
    for i in range(dataframe.shape[0]):
        instance = dataframe.iloc[i]["Instance"]
        model = dataframe.iloc[i]["Model"]
        X = dataframe.iloc[i]["X"]
        Y = dataframe.iloc[i]["Y"]
        R_sq = dataframe.iloc[i]["R_sq_log"]

        print(instance + " - " + model + ": " + str(dataframe.iloc[i]["Score"]))
        plt.scatter(
            X[:, 0], X[:, 1], c=np.arctan(X[:, 0]), cmap="rainbow", s=25, alpha=0.8
        )
        plt.savefig("results/images/" + instance + "_" + model + "_grid.png")
        plt.show()

        plt.scatter(
            Y[:, 0], Y[:, 1], c=np.arctan(X[:, 0]), cmap="rainbow", s=25, alpha=0.8
        )
        plt.savefig("results/images/" + instance + "_" + model + "_image.png")
        plt.show()

        plt.scatter(X[:, 0], X[:, 1], c=R_sq, cmap="gray", s=25)
        plt.savefig(
            "results/images/" + instance + "_" + model + "_curvature_grid_gray.png"
        )
        plt.show()

        plt.scatter(Y[:, 0], Y[:, 1], c=R_sq, cmap="gray", s=25)
        plt.savefig(
            "results/images/" + instance + "_" + model + "_curvature_image_gray.png"
        )
        plt.show()

        plt.scatter(X[:, 0], X[:, 1], c=R_sq, cmap="OrRd", s=25)
        plt.savefig("results/images/" + instance + "_" + model + "_curvature_grid.png")
        plt.show()

        plt.scatter(Y[:, 0], Y[:, 1], c=R_sq, cmap="OrRd", s=25)
        plt.savefig("results/images/" + instance + "_" + model + "_curvature_image.png")
        plt.show()


# -

if __name__ == "__main__":

    def eval_pair(y_true, y_pred):
        os.system(
            "taskset -cp 0-%d %s > /dev/null 2>&1"
            % (multiprocessing.cpu_count(), os.getpid())
        )
        return L2_norm_sectional_curvature(
            y_true,
            y_pred,
            metric_estimation="interpolate_metric",
            verbose=False,
            full_output=True,
        )

    with open("results/projection_dataset_names.pickle", "rb") as file:
        projection_names = pickle.load(file)
    projection_results = np.load("results/projection_results.npy")

    with parallel_backend("loky"):
        print("Evaluating")
        start_time = datetime.datetime.now()
        print(f"Start time: {start_time}")
        curvatures = Parallel(n_jobs=-1)(
            delayed(eval_pair)(x, y)
            for x, y in zip([GRID] * len(projection_results), projection_results)
        )

        results = [
            (*name.split(" "), curvature[0], curvature[1], curvature[2], curvature[3])
            for name, curvature in zip(projection_names, curvatures)
        ]

        df_result = pd.DataFrame(
            [results[i] for i in range(len(results))],
            columns=["Instance", "Model", "X", "Y", "R_sq_log", "Score"],
            index=None,
        )

        end_time = datetime.datetime.now()
        print(f"End time: {end_time}")
        print(f"Time elapsed: {end_time - start_time}")

        df_result.to_csv("results/results.csv")
        plot_results(df_result)



import multiprocessing
import os

import psutil

from definitions import TARGET_DIMENSIONALITY


def setup_multiprocessing():
    process = psutil.Process(os.getpid())
    process.cpu_affinity(range(multiprocessing.cpu_count()))


def get_instance_name(instance):
    name = instance[0][0].upper() + instance[1][0].upper()
    name += (
        ("E" if instance[2] == 0.8 else "H")
        + ("E" if instance[3] == 0.8 else "H")
        + str(TARGET_DIMENSIONALITY)
    )
    return name



import multiprocessing
import os
import pickle
from itertools import product

import numpy as np
from joblib import Parallel, delayed, parallel_backend
from scipy.integrate import quad
from scipy.interpolate import CubicSpline
from scipy.stats import multivariate_normal
from scipy.stats import special_ortho_group

from definitions import (
    RANDOM_NOISE_STD,
    GRID,
    TARGET_DIMENSIONALITY,
    USED_CURVATURES,
    CURVATURES,
    DIFFICULTY,
    RANDOM_SEED,
)
from utils import get_instance_name, setup_multiprocessing


def create_curve(
    curvature_function, x0=(0, 0), v0=(1, 1), t_min=-1, t_max=1, t_step=0.01
):
    """
    Returns a function that will apply a single curvature to its arguments.
    :param curvature_function: callable. Will apply the curvature itself.
    # TODO: explanations for the rest
    :param x0:
    :param v0:
    :param t_min:
    :param t_max:
    :param t_step:
    """
    t_num = int((t_max - t_min) / t_step)
    xis = np.linspace(t_min, t_max, 10 * t_num)
    theta_fun = np.vectorize(lambda x: quad(curvature_function, 0, x)[0])
    theta_val = theta_fun(xis)
    sp_theta = CubicSpline(xis, theta_val)

    def _tang_x(xi):
        thet = sp_theta(xi)
        return v0[0] * np.cos(thet) - v0[1] * np.sin(thet)

    def _tang_y(xi):
        thet = sp_theta(xi)
        return v0[0] * np.sin(thet) + v0[1] * np.cos(thet)

    si = np.linspace(t_min, t_max, 10 * t_num)

    res_x_fun = np.vectorize(lambda x: quad(_tang_x, 0, x)[0] + x0[0])
    res_y_fun = np.vectorize(lambda x: quad(_tang_y, 0, x)[0] + x0[1])

    res_x_val = res_x_fun(si)
    res_y_val = res_y_fun(si)

    sp_res_x = CubicSpline(si, res_x_val)
    sp_res_y = CubicSpline(si, res_y_val)

    def _gamma(s):
        return np.array([sp_res_x(s), sp_res_y(s)])

    return _gamma


def get_curvature_function(curvature_types, args, n, t_min=-1, t_max=1, t_step=0.01):
    """
    Generates a curvature function, that will apply a number of curvatures to its arguments.
    :param curvature_types: list of str containing keys of CURVATURES. Individual curvature functions to be used.
    :param args: list of arguments for every curvature function.
    :param n: target dimensionality.
    #TODO: t_min, t_max, t_step
    :param t_min:
    :param t_max:
    :param t_step:
    """
    curvatures = [
        CURVATURES[curvature_types[i]](args[i]) for i in range(len(curvature_types))
    ]

    individual_curvature_functions = [
        create_curve(curvature, t_min=t_min, t_max=t_max, t_step=t_step)
        for curvature in curvatures
    ]

    def zero_pad(arg, start, end=n):
        return np.array([0] * start + list(arg) + [0] * (end - start - 2))

    def _X(arg):
        return sum(
            [
                zero_pad(individual_curvature_functions[i](arg[i]), i)
                for i in range(len(curvatures))
            ]
        )

    return _X


def create_dataset(curvature_types, args, n, std):
    """
    Generates a Phi function that applies a number of curvatures to its arguments,
    increasing its dimensionality to n.
    The result will then be rotated randomly and applied a random translation.
    :param curvature_types: a list of str with the curvatures to use. Values in this list must be found in CURVATURES.
    :param args: a list of arguments to supply for every curvature type.
    :param n: target dimensionality.
    :param std: standard deviation to apply to the random_translation.
    """
    if len(curvature_types) > n - 1:
        raise ValueError(
            "Too many curvature functions have been supplied. "
            "The resulting manifold dimensionality would exceed the target dimensionality."
        )

    curvature_function = get_curvature_function(curvature_types, args, n)
    covariance = np.eye(n) * std ** 2

    def phi(x):
        random_rotation_matrix = special_ortho_group.rvs(n, random_state=RANDOM_SEED)
        random_translation = multivariate_normal.rvs(
            mean=np.zeros(n), cov=covariance, random_state=RANDOM_SEED
        )
        return np.dot(
            random_rotation_matrix, curvature_function(x) + random_translation.T
        ).T

    return phi


def gen_pair(instance):
    os.system(
        "taskset -cp 0-%d %s > /dev/null 2>&1"
        % (multiprocessing.cpu_count(), os.getpid())
    )
    name = get_instance_name(instance)
    return name, np.apply_along_axis(
        create_dataset(
            [instance[0], instance[1]],
            (instance[2], instance[3]),
            TARGET_DIMENSIONALITY,
            RANDOM_NOISE_STD,
        ),
        1,
        GRID,
    )


if __name__ == "__main__":

    setup_multiprocessing()

    with parallel_backend("loky"):
        print("Generating problem instances...")
        results = Parallel(n_jobs=-1)(
            delayed(gen_pair)(x)
            for x in product(USED_CURVATURES, USED_CURVATURES, DIFFICULTY, DIFFICULTY)
        )
        names = [x[0] for x in results]
        data = np.stack([x[1] for x in results])
        with open("results/dataset_names.pickle", "wb") as file:
            pickle.dump(names, file)
        np.save("results/dataset.npy", data)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['manifoldy', 'manifoldy.test']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18', 'scipy>=1.4', 'ndsplines>=0.1.2', 'joblib>=1.0.1']

extras_require = \
{'doc': ['sphinx'],
 'test': ['pytest >=2.7.3',
          'pytest-cov',
          'numpy>=1.18',
          'scipy>=1.4',
          'ndsplines>=0.1.2',
          'joblib>=1.0.1']}

setup(name='manifoldy',
      version='0.5.0',
      description='A package to generate customizable manifold synthetic datasets.',
      author='Raul Lara-Cabrera',
      author_email='raul.lara@upm.es',
      url='https://github.com/KNODIS-Research-Group/manifoldy',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.8',
     )



import numpy as np
from sklearn.decomposition import PCA, TruncatedSVD

from sklearn.manifold import Isomap
from sklearn.manifold import TSNE
from umap import UMAP

# GENERAL PARAMETERS
RANDOM_SEED = 42
MULTIPROCESSING_POOL_SIZE = 40
GRID = np.mgrid[0:1:30j, 0:1:30j].reshape(2, -1).T

# DIMENSIONALITY AUGMENTATION PARAMETERS
TARGET_DIMENSIONALITY = 7
RANDOM_NOISE_STD = 0.01
USED_CURVATURES = ("logistic", "polynomial_roll", "sine", "circle", "flat")
CURVATURES = {
    "flat": lambda x: lambda t: 0,
    "circle": lambda x: lambda t: x * np.pi * 2,
    "polynomial_roll": lambda x: lambda t: x * 4 * (t + 1) ** (x * 2),
    "roll": lambda x: lambda t: np.exp(x * t * 4),
    "gaussian": lambda x: lambda t: 1 / (x * 0.1) * np.exp(-(t ** 2) / (x * 0.1) ** 2),
    "sine": lambda x: lambda t: (5 + (x - 1) * 10) * np.sin(t * np.pi * 2),
    "logistic": lambda x: lambda t: (x * 10) / (1 + np.exp(-0.5 * t)),
    "relu": lambda x: lambda t: 0 if t < 0 else x * t * 10,
}

# DIMENSIONALITY REDUCTION PARAMETERS
DIFFICULTY = (0.8, 1.2)
DIMENSIONALITY_REDUCTION_MODELS = (
    PCA(n_components=2, random_state=RANDOM_SEED),
    Isomap(n_components=2, eigen_solver="dense"),
    TruncatedSVD(n_components=2, random_state=RANDOM_SEED),
    TSNE(n_components=2, random_state=RANDOM_SEED),
    UMAP(n_components=2, random_state=RANDOM_SEED, metric="cosine"),
)



# -*- coding: utf-8 -*-
# coding: utf-8

import copy
import ndsplines
import numpy as np
from scipy.special import comb
from scipy import integrate
from scipy.interpolate import griddata

# COMPUTATION OF CHRISTOFFEL SYMBOLS AND CURVATURE


def compute_first_christoffel_symbols(dg, n):
    CS_first = np.zeros((n, n, n))
    for i in range(n):
        for j in range(n):
            for k in range(n):
                CS_first[k, i, j] = 1 / 2 * (dg[j][k, i] + dg[i][k, j] - dg[k][i, j])

    return CS_first


def compute_partial_derivative_first_christoffel_symbols(ddg, l, n):
    CS_first_diff = np.zeros((n, n, n))
    for i in range(n):
        for j in range(n):
            for k in range(n):
                CS_first_diff[k, i, j] = (
                    1 / 2 * (ddg[l][j][k, i] + ddg[l][i][k, j] - ddg[l][k][i, j])
                )

    return CS_first_diff


def compute_derivative_first_christoffel_symbols(ddg, n):
    derivatives = []
    for l in range(n):
        CS_first_diff = compute_partial_derivative_first_christoffel_symbols(ddg, l, n)
        derivatives.append(CS_first_diff)
    return np.stack(derivatives)


def compute_christoffel_symbols(g, dg):
    # Index 0: Upper index
    # Index 1, 2: Lower indices
    # Gamma_{1,2}^0
    n = g.shape[0]
    CS_first = compute_first_christoffel_symbols(dg, n)

    g_inv = np.linalg.inv(g)
    return np.tensordot(g_inv, CS_first, axes=([1], [0]))


def compute_derivative_christoffel_symbols(g, dg, ddg):
    # Index 0: Derivative of the symbol
    # Index 1: Upper index
    # Index 2, 3: Lower indices
    # Gamma_{0;2,3}^1
    n = g.shape[0]
    CS_first = compute_first_christoffel_symbols(dg, n)
    g_inv = np.linalg.inv(g)

    derivatives = []
    for l in range(n):
        CS_first_diff_l = compute_partial_derivative_first_christoffel_symbols(
            ddg, l, n
        )
        g_inv_diff_l = -np.matmul(np.matmul(g_inv, dg[l]), g_inv)
        CS_l = np.tensordot(g_inv_diff_l, CS_first, axes=([1], [0])) + np.tensordot(
            g_inv, CS_first_diff_l, axes=([1], [0])
        )
        derivatives.append(CS_l)

    return np.stack(derivatives)


def compute_curvature_tensor(g, dg, ddg):
    # Index 0: Upper index
    # Index 1, 2, 3: Lower indices
    # R[l,i,j,k] = R^l_{ijk} = dx^l(R(ei,ej)ek)
    n = g.shape[0]

    CS = compute_christoffel_symbols(g, dg)
    CS_diff = compute_derivative_christoffel_symbols(g, dg, ddg)
    R = np.zeros((n, n, n, n))
    CS_contract = np.tensordot(CS, CS, axes=([0], [2]))

    for i in range(n):
        for j in range(n):
            for k in range(n):
                for l in range(n):
                    R[l, i, j, k] = (
                        CS_diff[j, l, i, k]
                        - CS_diff[i, l, j, k]
                        + CS_contract[i, k, l, j]
                        - CS_contract[j, k, l, i]
                    )

    return R


def compute_first_curvature_tensor(g, dg, ddg):
    # Index 0: Upper index
    # Index 1, 2, 3: Lower indices
    # R[l,i,j,k] = R^l_{ijk} = dx^l(R(ei,ej)ek)
    # n = g.shape[0]

    R = compute_curvature_tensor(g, dg, ddg)
    R_first = np.tensordot(R, g, axes=([0], [0]))

    return R_first


def compute_sectional_curvatures(g, dg, ddg):
    # Returns a vector whose entries are
    # sectional curvatures at (e1,e2), (e1,e3),...,(e1,en)
    # (e2,e3), (e2,e4), ..., (e2, en),...
    n = g.shape[0]
    R = compute_first_curvature_tensor(g, dg, ddg)
    sect_curvs = np.zeros(int(comb(n, 2)))
    counter = 0
    for i in range(n):
        for j in range(i + 1, n):
            area = g[i, i] * g[j, j] - g[i, j] ** 2
            sect_curvs[counter] = R[i, j, i, j] / area
            counter += 1

    return np.stack(sect_curvs)


# UTILS


def dataset_to_grid(X, n_samples, dim):
    shap = n_samples + [dim]
    # shap = [d] + n_samples
    return X.reshape(shap)


def grid_data(X, Y, n_samples=None):
    dim = X.shape[1]
    if n_samples is None:
        n_samples = [41] * dim

    offset = 0.05
    xs = np.array(
        [np.linspace(0 + offset, 1 - offset, n_samples[i]) for i in range(dim)]
    )
    X_new = np.array(np.meshgrid(*xs))

    Y_new = []
    for i in range(dim):
        grid_y = griddata(X, Y[:, i], tuple(X_new), method="linear")
        Y_new.append(grid_y)
    Y_new = np.stack(Y_new)

    n_total_samples = 1
    for n in n_samples:
        n_total_samples *= n

    return (
        X_new.reshape(dim, n_total_samples).T,
        Y_new.reshape(dim, n_total_samples).T,
        n_samples,
    )


def estimate_inner_product(X, b):
    u, s, vh = np.linalg.svd(X, full_matrices=True)
    S = np.zeros((u.shape[0], vh.shape[0]))
    S_inv = np.zeros((vh.shape[0], u.shape[0]))
    for i, e in enumerate(s):
        S[i, i] = e
        S_inv[i, i] = 1 / e

    return vh.T @ (S_inv @ (u.T @ (b @ (u @ (S_inv.T @ vh)))))


def estimate_tensor_left(X, b, arity):
    X_inv = np.linalg.pinv(X)
    result = b
    for _ in range(arity):
        result = np.tensordot(X_inv, result, axes=([1], [0]))

    return result


def quadrature_regular_grid(X, values, grid_n_samples):
    # d = len(grid_n_samples)
    x0 = X[0]
    index = 1
    volume = 1
    n_samples = copy.copy(grid_n_samples)
    aux = n_samples[0]
    n_samples[0] = n_samples[1]
    n_samples[1] = aux
    n_samples = [1] + n_samples[:-1]
    for i in n_samples:
        index = index * i
        volume = volume * np.linalg.norm(X[index] - x0)

    return np.sum(values, axis=0) * volume


def crop_matrix(X, crop_indices):
    if len(X.shape) == 1:
        return X[crop_indices[0]]

    return np.stack([crop_matrix(f, crop_indices[1:]) for f in X[crop_indices[0]]])


# ESTIMATION OF THE RIEMANNIAN METRIC


def get_indexes_best_neighbors(X, K, i, d):
    index_neig = [-1] * K
    dist_neig = np.array([np.inf] * K)

    for j in range(X.shape[0]):
        if d(X[i], X[j]) < np.max(dist_neig) and j != i:
            index_max = np.argmax(dist_neig)
            index_neig[index_max] = j
            dist_neig[index_max] = d(X[i], X[j])

    return index_neig


def get_riemannian_metric(X, Y, K):
    d = lambda x, y: np.linalg.norm(x - y)

    g = []
    dim = X.shape[1]
    for i in range(X.shape[0]):
        stop = False
        k = K
        while not stop:
            index_neig = get_indexes_best_neighbors(X, k, i, d)
            X_extract = X[index_neig]
            X_vec = X_extract - X[i]
            if np.linalg.matrix_rank(X_vec) >= dim:
                stop = True
            else:
                k += 1

        Y_extract = Y[index_neig]
        Y_vec = Y_extract - Y[i]
        b = np.matmul(Y_vec, Y_vec.T)

        g.append(estimate_inner_product(X_vec, b))
    g = np.stack(g)
    return g


def compute_derivative_point(X, Y, i, K):
    d = lambda x, y: np.linalg.norm(x - y)

    stop = False
    k = K
    dim = X.shape[1]
    while not stop:
        index_neig = get_indexes_best_neighbors(X, k, i, d)
        X_extract = X[index_neig]
        X_vec = X_extract - X[i]
        if np.linalg.matrix_rank(X_vec) >= dim:
            stop = True
        else:
            k += 1

    # new_shape = (k,) + Y.shape[1:]
    # b = np.zeros(new_shape)
    Y_extract = Y[index_neig]
    Y_vec = Y_extract - Y[i]

    return estimate_tensor_left(X_vec, Y_vec, 1)


def compute_derivative(X, Y, K):
    der = []
    for i in range(X.shape[0]):
        der.append(compute_derivative_point(X, Y, i, K))

    return np.stack(der)


def compute_derivative_function(X, Y, K):
    df = compute_derivative(X, Y, K)
    ddf = compute_derivative(X, df, K)
    dddf = compute_derivative(X, ddf, K)

    return Y, df, ddf, dddf


def interpolate_derivative_function(X, Y):
    dim = Y.shape[1]
    n = X.shape[0]

    f = []
    for i in range(dim):
        tidy_data = np.hstack((X, Y[:, i].reshape(n, 1)))
        interp = ndsplines.make_interp_spline_from_tidy(
            tidy_data,
            range(dim),  # columns to use as independent variable data
            [dim],  # columns to use as dependent variable data
        )
        f.append(interp)

    derivs, dderivs, ddderivs = compute_derivatives_dirs(dim, 3)

    # DERIVATIVE
    df = []
    for der in derivs:
        dfj = []
        for j in range(dim):
            dfj.append(f[j](X, nus=der))
        df.append(np.stack(dfj))
    df = np.stack(df)
    df = np.moveaxis(df.reshape(dim, dim, n), [0, 1, 2], [1, 2, 0])

    # SECOND DERIVATIVES (HESSIAN)
    ddf = []
    for der in dderivs:
        ddfj = []
        for j in range(dim):
            ddfj.append(f[j](X, nus=der))
        ddf.append(np.stack(ddfj))
    ddf = np.stack(ddf)
    ddf = np.moveaxis(ddf.reshape(dim, dim, dim, n), [0, 1, 2, 3], [1, 2, 3, 0])

    # THIRD DERIVATIVES
    dddf = []
    for der in ddderivs:
        dddfj = []
        for j in range(dim):
            dddfj.append(f[j](X, nus=der))
        dddf.append(np.stack(dddfj))
    dddf = np.stack(dddf)
    dddf = np.moveaxis(
        dddf.reshape(dim, dim, dim, dim, n), [0, 1, 2, 3, 4], [1, 2, 3, 4, 0]
    )

    return Y, df, ddf, dddf


def interpolate_metric_derivatives(g_sp, X, n_samples):
    n = 1
    for n_s in n_samples:
        n = n * n_s
    d = len(n_samples)
    grid = dataset_to_grid(X, n_samples, X.shape[1])
    g = np.array([g_sp[i](grid) for i in range(d * d)])
    g = g.T.reshape((n, d, d))

    derivs, dderivs = compute_derivatives_dirs(d)

    dg = []
    for der in derivs:
        dg.append(np.array([g_sp[i](grid, nus=der) for i in range(d * d)]))

    dg = np.array(dg).reshape((d, d, d, n))
    dg = np.moveaxis(dg, [0, 1, 2, 3], [1, 2, 3, 0])

    ddg = []
    for der in dderivs:
        ddg.append(np.array([g_sp[i](grid, nus=der) for i in range(d * d)]))

    ddg = np.array(ddg).reshape((d, d, d, d, n))
    ddg = np.moveaxis(ddg, [0, 1, 2, 3, 4], [1, 2, 3, 4, 0])

    return g, dg, ddg


def interpolate_metric(X, Y, K):
    n = X.shape[0]
    d = X.shape[1]
    g = get_riemannian_metric(X, Y, K)
    inter = []
    for i in range(d):
        for j in range(d):
            tidy_data = np.hstack((X, g[:, i, j].reshape(n, 1)))
            interp = ndsplines.make_interp_spline_from_tidy(
                tidy_data,
                range(d),  # columns to use as independent variable data
                [d],  # columns to use as dependent variable data
            )
            inter.append(interp)

    return inter


def compute_derivatives_dirs(d, n_der=2):
    derivs = []
    for i in range(d):
        der = np.array([0] * d)
        der[i] = 1
        derivs.append(der)

    if n_der == 1:
        return derivs

    dderivs = []
    for i in derivs:
        for j in range(d):
            der = copy.deepcopy(i)
            der[j] = der[j] + 1
            dderivs.append(der)

    if n_der == 2:
        return derivs, dderivs

    ddderivs = []
    for i in dderivs:
        for j in range(d):
            der = copy.deepcopy(i)
            der[j] = der[j] + 1
            ddderivs.append(der)

    return derivs, dderivs, ddderivs


# TOP FUNCTIONS


def estimate_metric_derivatives(X, Y, K, grid_n_samples, metric_estimation="KNN"):
    if metric_estimation == "KNN":
        _, df, ddf, dddf = compute_derivative_function(X, Y, K)
    elif metric_estimation == "interpolate":
        _, df, ddf, dddf = interpolate_derivative_function(X, Y)
    elif metric_estimation == "interpolate_metric":
        g_sp = interpolate_metric(X, Y, K)
        g, dg, ddg = interpolate_metric_derivatives(g_sp, X, grid_n_samples)
        return g, dg, ddg

    g = np.einsum("ijs,iks->ijk", df, df)
    dg = np.einsum("ijks,ils->ijkl", ddf, df) + np.einsum("iks,ijls->ijkl", df, ddf)
    ddg = (
        np.einsum("ijkls,ims->ijklm", dddf, df)
        + np.einsum("ikls,ijms->ijklm", ddf, ddf)
        + np.einsum("ijls,ikms->ijklm", ddf, ddf)
        + np.einsum("ils,ijkms->ijklm", df, dddf)
    )

    return g, dg, ddg


def estimate_sectional_curvature(X, Y, K, grid_n_samples, metric_estimation="KNN"):
    n = X.shape[0]

    g, dg, ddg = estimate_metric_derivatives(X, Y, K, grid_n_samples, metric_estimation)

    R = []
    for i in range(n):
        R.append(compute_sectional_curvatures(g[i], dg[i], ddg[i]))
    R = np.stack(R)

    return R


def compute_norm(
    X, R_sq, grid_n_samples, offset, integration="adaptative", verbose=False
):
    dim = X.shape[1]
    if integration == "adaptative":
        result = 0
        for i in range(R_sq.shape[1]):
            tidy_data = np.hstack((X, R_sq[:, i].reshape(X.shape[0], 1)))
            interp = ndsplines.make_interp_spline_from_tidy(
                tidy_data,
                range(dim),  # columns to use as independent variable data
                [dim],  # columns to use as dependent variable data
            )

            def f(*argv):
                return interp(argv)[0, 0]

            if verbose:
                print("Initiating adaptative integration")
            result += integrate.nquad(
                f, [[offset, 1 - offset]] * dim, full_output=False
            )[0]
        return np.sqrt(result)

    reduction_indices = [
        range(int(grid_n_samples[i] * offset), int(grid_n_samples[i] * (1 - offset)))
        for i in range(len(grid_n_samples))
    ]
    X_grid = dataset_to_grid(X, grid_n_samples, dim)
    X_red_indices = reduction_indices + [range(dim)]
    X_crop_grid = crop_matrix(X_grid, X_red_indices)
    R_sq_grid = dataset_to_grid(R_sq, grid_n_samples, R_sq.shape[1])
    R_sq_red_indices = reduction_indices + [range(R_sq.shape[1])]
    R_sq_crop_grid = crop_matrix(R_sq_grid, R_sq_red_indices)

    n_crop = 1
    grid_n_samples_crop = [len(r) for r in reduction_indices]
    for r in grid_n_samples_crop:
        n_crop = n_crop * r
    X_crop = X_crop_grid.reshape(n_crop, X.shape[1])
    R_sq_crop = R_sq_crop_grid.reshape(n_crop, R_sq.shape[1])

    if verbose:
        print("Initiating grid integration")
    return np.sqrt(
        quadrature_regular_grid(X_crop, R_sq_crop, grid_n_samples_crop).sum()
    )


def L2_norm_sectional_curvature(
    X,
    Y,
    metric_estimation="KNN",
    K=20,
    integration="grid",
    grid_n_samples=None,
    offset=0.25,
    verbose=False,
):
    """
    Available metric_estimation:
        * 'KNN': Computes metric derivatives using K nearest neighbors.
        * 'interpolate': Interpolates the underlying function and uses it to compute the metric.
        * 'interpolate_metric': Computes the metric from the data and interpolate it to compute derivatives.

    Availabel integration:
        * 'grid': Computes the integral using a grid.
        * 'adaptative': Computes the integral using an adaptative method.
    """
    if grid_n_samples is None:
        X, Y, grid_n_samples = grid_data(X, Y)
    if verbose:
        print("Starting: estimation of sectional curvature")
    R = estimate_sectional_curvature(X, Y, K, grid_n_samples, metric_estimation)
    R_sq = R ** 2
    if verbose:
        print("Estimation of sectional curvature: DONE")

    return compute_norm(X, R_sq, grid_n_samples, offset, integration, verbose)
